import { Field } from 'snarkyjs';
export { MerkleTree };
export type { MerklePath, Options };
/**
 * A {@link MerklePath} ([or Witness, Merkle Proof , ..](https://computersciencewiki.org/index.php/Merkle_proof)) serves as a 'proof of inclusion'.
 * The data structure contains the path from a given leaf to the root of the Merkle Tree.
 */
declare type MerklePath = Array<{
    direction: Field;
    hash: Field;
}>;
/**
 * Option interface for a Merkle Tree
 */
interface Options {
    hashLeaves: boolean;
}
/**
 * A {@link MerkleTree} is a {@link BinaryTree} which aggregates hashes of the underlying data. See [Merkle Tree](https://en.wikipedia.org/wiki/Merkle_tree)
 *
 */
declare class MerkleTree {
    private tree;
    private options;
    /**
     * Builds a Merkle Tree based on a list of given leaves.
     * @param {Array<Field>} leaves leaf nodes
     * @param {Options} options that can define the structure of the merkle tree
     * @return {@link MerkleTree}
     */
    constructor(leaves: Array<Field>, options: Options);
    /**
     * Static function to validate a Merkle Path - Can be used within a zkApp/Circuit.
     * @param {MerklePath} merklePath path leading to the root of the tree
     * @param {Field} leafHash hash of element that needs verification
     * @param {Field} merkleRoot (target) root of the Merkle Tree
     * @returns {boolean}
     */
    static validateProof(merklePath: MerklePath, leafHash: Field, merkleRoot: Field): boolean;
    /**
     * Returns the Merkle Root
     * @returns {Field | undefined}
     */
    getMerkleRoot(): Field | undefined;
    /**
     * Returns a merkle path of an element at a given index
     * @param {number} index of element
     * @returns {MerklePath | undefined} merkle path or undefined
     */
    getProof(index: number): MerklePath;
    /**
     * Finds the index of a given element
     * @param {number} element to find
     * @returns {number | undefined} index or undefined
     */
    getIndex(element: Field): number | undefined;
    /**
     * Appends new leaves to an existing Merkle Tree
     * @param {Array<Field>} leaves
     * @param {boolean} hash if true elements in the array will be hased using Poseidon, if false they will be inserted directly
     */
    appendLeaves(leaves: Array<Field>, hash?: boolean): void;
    /**
     * (Re)builds the {@link MerkleTree}'s levels based on pre-initialized leaves
     */
    private makeTree;
    /**
     * Calculates new levels of the Merkle Tree structure, helper function
     * @returns {Array<Field>} Level of the Merkle Tree
     */
    private calculateNextLevel;
    printTree(): void;
    printProof(index: number): void;
}
